from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

import numpy as np

from stereomolgraph.coords import (
    BondsFromDistance,
    angle_from_coords,
    are_planar,
    handedness,
)
from stereomolgraph.stereodescriptors import (
    AtomStereo,
    Octahedral,
    PlanarBond,
    SquarePlanar,
    Tetrahedral,
    TrigonalBipyramidal,
)

if TYPE_CHECKING:
    from typing import Literal, TypeVar

    from stereomolgraph import MolGraph, StereoMolGraph
    from stereomolgraph.coords import BondsFromDistance, Geometry

    N = TypeVar("N", bound=int, covariant=True)
    MG = TypeVar("MG", bound=MolGraph, covariant=True)
    SMG = TypeVar("SMG", bound=StereoMolGraph, covariant=True)
    NP_FLOAT = TypeVar(
        "NP_FLOAT", bound=np.dtype[np.floating], contravariant=True
    )

    THREE = Literal[3]


def connectivity_from_geometry(
    cls: type[MG],
    geo: Geometry,
    switching_function: BondsFromDistance = BondsFromDistance(),
) -> MG:
    """
    Creates a graph of a molecule from a Geometry and a switching Function.
    Uses the Default switching function if none are given.

    :param geo: Geometry
    :param switching_function: Function to determine if two atoms are
        connected
    :return: graph of molecule
    """

    connectivity_matrix = switching_function.array(geo.coords, geo.atom_types)
    return cls.from_atom_types_and_bond_order_matrix(
        geo.atom_types,
        connectivity_matrix,
    )


def stero_from_geometry(
    smg: SMG,
    geo: Geometry,
) -> SMG:
    for atom in range(geo.n_atoms):
        first_nbrs = smg.bonded_to(atom)
        atom_stereo_tup = (atom, *first_nbrs)
        atom_stereo = atom_stereo_from_coords(
            atom_stereo_tup, geo.coords.take(atom_stereo_tup, axis=0)
        )

        if atom_stereo is not None:
            smg.set_atom_stereo(atom_stereo)

        if len(first_nbrs) != 3:
            continue

        for nbr in first_nbrs:
            second_neighbors = smg.bonded_to(nbr).difference({atom})

            if len(second_neighbors) != 2:
                continue

            first_nbrs_reduced = first_nbrs.difference({nbr})
            stereo_atoms = (*first_nbrs_reduced, atom, nbr, *second_neighbors)
            assert len(stereo_atoms) == 6
            planar_bond = _planar_bond_from_coords(
                stereo_atoms, geo.coords.take(stereo_atoms, axis=0)
            )
            if planar_bond is not None:
                smg.set_bond_stereo(planar_bond)

    return smg


def atom_stereo_from_coords(
    atoms: tuple[int, ...],
    coords: np.ndarray[tuple[N, THREE], NP_FLOAT],
) -> None | AtomStereo:
    assert len(atoms) == coords.shape[0]
    assert coords.shape[1] == 3
    assert np.issubdtype(coords.dtype, np.floating)

    if len(atoms) == 5 and coords.shape[0] == 5:
        if are_planar(coords[[1, 2, 3, 4]]):
            return _square_planar_from_coords(atoms=atoms, coords=coords)  # type: ignore
        else:
            return _tetrahedral_from_coords(atoms=atoms, coords=coords)  # type: ignore

    elif len(atoms) == 6:
        return _trigonal_bipyramidal_from_coords(atoms=atoms, coords=coords)  # type: ignore

    elif len(atoms) == 7:
        return _octahedral_from_coords(atoms=atoms, coords=coords)  # type: ignore

    else:
        return None


def _tetrahedral_from_coords(
    atoms: tuple[int, int, int, int, int],
    coords: np.ndarray[tuple[Literal[5], THREE], NP_FLOAT],
) -> Tetrahedral:
    """
    Creates the representation of a Tetrahedral Stereochemistry
    from the coordinates of the atoms.

    :param atoms: Atoms of the stereochemistry
    :param coords: nAtomsx3 numpy array with cartesian coordinates
    """
    indeces = (1, 2, 3, 4)
    orientation = handedness(coords.take(indeces, axis=0))
    int_orientation = int(orientation)
    assert int_orientation in (1, -1), (
        f"Orientation {orientation} is not valid for Tetrahedral "
        "stereochemistry."
    )
    return Tetrahedral(atoms, int_orientation)


def _square_planar_from_coords(
    atoms: tuple[int, int, int, int, int],
    coords: np.ndarray[tuple[Literal[5], THREE], NP_FLOAT],
) -> None | SquarePlanar:
    if not are_planar(coords[[1, 2, 3, 4]]):
        return None

    distinct_orders = ((1, 2, 3, 4), (2, 1, 3, 4), (3, 2, 1, 4), (4, 2, 3, 1))

    max_order: tuple[int, int, int, int] = (1, 2, 3, 4)
    angle_sum = 0.0
    for order in distinct_orders:
        a = angle_from_coords(coords[[order[i] for i in (0, 1, 2)]])
        b = angle_from_coords(coords[[order[i] for i in (1, 2, 3)]])
        c = angle_from_coords(coords[[order[i] for i in (2, 3, 0)]])
        d = angle_from_coords(coords[[order[i] for i in (3, 0, 1)]])
        angle = a + b + c + d
        if angle > angle_sum:
            angle_sum, max_order = angle, order

    atoms = (
        atoms[0],
        atoms[max_order[0]],
        atoms[max_order[1]],
        atoms[max_order[2]],
        atoms[max_order[3]],
    )

    return SquarePlanar(atoms, parity=0)


def _trigonal_bipyramidal_from_coords(
    atoms: tuple[int, int, int, int, int, int],
    coords: np.ndarray[tuple[Literal[6], THREE], NP_FLOAT],
) -> None | TrigonalBipyramidal:
    """
    calculates the distance of the atom 5 from the plane defined by the
    first three atoms in Angstrom. The sign of the distance is determined
    by the side of the plane that atom 5 is on.
    """
    indices = (1, 2, 3, 4, 5)

    if np.any(are_planar(coords[[1, 2, 3, 4, 5]])):
        return None

    lst = np.array(
        [[i, 0, j] for i, j in itertools.combinations(indices, 2)],
        dtype=np.int8)

    # The atoms with the largest angle are the axial atoms
    angles = angle_from_coords(coords[lst])

    i, j = lst[angles.argmax()][[0, 2]]  # axial atoms
    i, j = int(i), int(j)

    equatorial = [a for a in indices if a not in (i, j)]
    i_rotation = -1 * handedness(coords.take([*equatorial, i], axis=0))
    j_rotation = handedness(coords.take([*equatorial, j], axis=0))

    assert int(i_rotation) == int(j_rotation)

    indeces_in_new_order = (i, j, *equatorial)

    orientation = int(i_rotation)
    tb_indeces = (0, *indeces_in_new_order)

    tb_atoms = tuple(atoms[i] for i in tb_indeces)

    assert len(tb_atoms) == 6
    assert orientation in (1, -1)
    return TrigonalBipyramidal(tb_atoms, orientation)


def _octahedral_from_coords(
    atoms: tuple[int, int, int, int, int, int, int],
    coords: np.ndarray[tuple[Literal[7], THREE], NP_FLOAT],
) -> None | Octahedral:
    indeces = (1, 2, 3, 4, 5, 6)
    planar_groups: list[set[int]] = []
    for p1, p2, p3, p4 in itertools.combinations(indeces, 4):
        points = [p1, p2, p3, p4]
        if are_planar(coords[points]):
            planar_groups.append(set(points))

    if not len(planar_groups) == 3:
        return None

    trans_atoms = planar_groups[0].intersection(planar_groups[1])

    cis_atoms0 = planar_groups[0].difference(trans_atoms)

    cis_atoms1 = planar_groups[1].difference(trans_atoms)
    assert len(trans_atoms) == 2
    assert len(cis_atoms0) == 2
    assert len(cis_atoms1) == 2
    assert cis_atoms0 | cis_atoms1 == planar_groups[2]

    a1, a2 = trans_atoms
    a3, a5 = cis_atoms0
    a4, a6 = cis_atoms1

    parity = int(handedness(coords[[a1, a3, a5, a4]]))
    assert parity == 1 or parity == -1
    return Octahedral((atoms[0], a1, a2, a3, a4, a5, a6), parity)


def _planar_bond_from_coords(
    atoms: tuple[int, int, int, int, int, int],
    coords: np.ndarray[tuple[Literal[6], Literal[3]], np.dtype[np.floating]],
) -> None | PlanarBond:
    if not are_planar(coords):
        return None

    a = (coords[0] - coords[1]) / np.linalg.norm(coords[0] - coords[1])
    b = (coords[4] - coords[5]) / np.linalg.norm(coords[4] - coords[5])
    result = int(np.sign(np.dot(a, b)))

    if result == -1:
        new_atoms = tuple(atoms[i] for i in (1, 0, 2, 3, 4, 5))
    elif result == 1:
        new_atoms = atoms
    elif result == 0:
        raise ValueError("atoms are tetrahedral")
    else:
        raise ValueError("something went wrong")
    assert len(new_atoms) == 6
    return PlanarBond(new_atoms, 0)



# pyright: standard
# typing with rdkit is not fully supported
from __future__ import annotations

from dataclasses import dataclass
from types import MappingProxyType
from typing import TYPE_CHECKING

import rdkit.Chem as Chem  # type: ignore

from stereomolgraph import AtomId, MolGraph, StereoMolGraph
from stereomolgraph.stereodescriptors import (
    AtomStereo,
    AtropBond,
    Octahedral,
    PlanarBond,
    SquarePlanar,
    Tetrahedral,
    TrigonalBipyramidal,
    BondStereo
)

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import ClassVar, Literal


def mol_graph_from_rdmol(
    cls: type[MolGraph], rdmol: Chem.Mol, use_atom_map_number: bool = False
) -> MolGraph:
    """
    Creates a StereoMolGraph from an RDKit Mol object.
    Implicit Hydrogens are added to the graph.
    Stereo information is conserved. Double bonds, aromatic bonds and
    conjugated bonds are interpreted as planar. Atoms with 5 bonding
    partners are assumed to be TrigonalBipyramidal and allow interchange
    of the substituents (berry pseudorotation). Atoms with 6 bonding
    partners are assumed to be octahedral and do not allow interchange of
    the substituents.

    :param rdmol: RDKit Mol object
    :param use_atom_map_number: If the atom map number should be used
                                instead of the atom index
    :return: StereoMolGraph
    """

    if use_atom_map_number is False:
        rdmol = Chem.rdmolops.AddHs(rdmol, explicitOnly=True)

    graph = cls()

    if use_atom_map_number:
        id_atom_map = {
            atom.GetIdx(): atom.GetAtomMapNum() for atom in rdmol.GetAtoms()
        }
    else:
        id_atom_map = {
            atom.GetIdx(): atom.GetIdx() for atom in rdmol.GetAtoms()
        }

    for atom in rdmol.GetAtoms():
        graph.add_atom(id_atom_map[atom.GetIdx()], atom.GetSymbol())

    for bond in rdmol.GetBonds():
        graph.add_bond(
            id_atom_map[bond.GetBeginAtomIdx()],
            id_atom_map[bond.GetEndAtomIdx()],
        )
    return graph


@dataclass
class RDMol2StereoMolGraph:
    """Convert an RDKit :class:`rdkit.Chem.Mol` to a :class:`StereoMolGraph`.

    All aromatic bonds are considered to be cis. Double bonds in rings are
    assumed to be cis for rings of size <= 7.

    :param stereo_complete: If ``True``, attempt to infer complete stereo
        parities when possible. Defaults to ``False``.
    :param use_atom_map_number: Use RDKit atom map numbers as atom identifiers
        instead of RDKit atom indices. Defaults to ``False``.
    :param lone_pair_stereo: Include stereochemistry that depends on lone
        pairs (if present). Defaults to ``True``.
    :param resonance: Enumerate resonance structures and merge bond stereo
        information from them. Defaults to ``True``.
    """
    stereo_complete: bool = False
    use_atom_map_number: bool = False
    lone_pair_stereo: bool = True
    resonance: bool = True
    _max_resonance_structures: int = 100
    _min_trans_ring_size: int = 7

    def __call__(self, rdmol: Chem.Mol) -> StereoMolGraph:
        smg = self.smg_from_rdmol(rdmol)

        if self.resonance is False:
            return smg
        
        elif self.resonance is True:
            enumerator = (res_mol for res_mol
                          in Chem.ResonanceMolSupplier(
                              rdmol, Chem.KEKULE_ALL,
                              self._max_resonance_structures)
                         if res_mol is not None)

            for res_mol in enumerator:
                res_smg = self.smg_from_rdmol(res_mol)
                for bond, bond_stereo in res_smg.bond_stereo.items():
                    if bond not in smg.bond_stereo:
                        smg.set_bond_stereo(bond_stereo)
        return smg

    def smg_from_rdmol(self, rdmol: Chem.Mol) -> StereoMolGraph:
        if rdmol is None:
            raise ValueError("rdmol is None")
        rdmol = Chem.AddHs(rdmol, explicitOnly=False)
        graph = StereoMolGraph()

        id_atom_map: dict[int, AtomId]

        if self.use_atom_map_number is True:
            if any(atom.GetAtomMapNum() == 0 for atom in rdmol.GetAtoms()):
                raise ValueError("AtomMapNumber has to  be set on all atoms")
            id_atom_map = {
                atom.GetIdx(): atom.GetAtomMapNum()
                for atom in rdmol.GetAtoms()
            }
        else:
            id_atom_map = {
                atom.GetIdx(): atom.GetIdx() for atom in rdmol.GetAtoms()
            }

        for atom in rdmol.GetAtoms():
            graph.add_atom(id_atom_map[atom.GetIdx()], atom.GetSymbol())

        for bond in rdmol.GetBonds():
            graph.add_bond(
                id_atom_map[bond.GetBeginAtomIdx()],
                id_atom_map[bond.GetEndAtomIdx()],
            )

        for atom in rdmol.GetAtoms():
            atom_idx: int = atom.GetIdx()

            neighbors: tuple[int, ...] = tuple(
                id_atom_map[n.GetIdx()]
                for n in rdmol.GetAtomWithIdx(atom_idx).GetNeighbors()
            )

            chiral_tag = atom.GetChiralTag()
            hybridization = atom.GetHybridization()

            atom_stereo: None | AtomStereo = None

            if chiral_tag in self._rd_tetrahedral:
                stereo_atoms = (id_atom_map[atom_idx], *neighbors)
                if len(stereo_atoms) == 5:
                    atom_stereo = Tetrahedral(
                        stereo_atoms,
                        self._rd_tetrahedral[chiral_tag],
                    )
                elif len(stereo_atoms) == 4:
                    atom_stereo = Tetrahedral(
                        (*stereo_atoms, None),
                        self._rd_tetrahedral[chiral_tag],
                    )
                else:
                    raise RuntimeError(
                        "Tetrahedral stereo must have 3 or 4 neighbors"
                    )

            elif chiral_tag == Chem.ChiralType.CHI_TETRAHEDRAL or (
                hybridization == Chem.HybridizationType.SP3
                and len(neighbors) == 4
            ):
                short_stereo_atoms = (id_atom_map[atom_idx], *neighbors)
                stereo_atoms = tuple(
                    short_stereo_atoms[i]
                    if i < len(short_stereo_atoms)
                    else None
                    for i in range(5)
                )  # extends with "None" if less than 4 neighbors
                assert len(stereo_atoms) == 5

                if not self.stereo_complete:
                    atom_stereo = Tetrahedral(stereo_atoms, None)
                elif self.stereo_complete:
                    atom_stereo = Tetrahedral(stereo_atoms, parity=1)
                else:
                    raise RuntimeError("This should never happen")

            elif chiral_tag == Chem.ChiralType.CHI_SQUAREPLANAR:
                sp_order: tuple[int, int, int, int]
                if atom.GetUnsignedProp("_chiralPermutation") == 1:
                    sp_order = (0, 1, 2, 3)
                elif atom.GetUnsignedProp("_chiralPermutation") == 2:
                    sp_order = (0, 2, 1, 3)
                elif atom.GetUnsignedProp("_chiralPermutation") == 3:
                    sp_order = (0, 1, 3, 2)
                else:
                    raise RuntimeError("Unknown permutation for SquarePlanar")
                ordered_neighbors = tuple([neighbors[i] for i in sp_order])
                sp_atoms = (id_atom_map[atom_idx], *ordered_neighbors)
                assert len(sp_atoms) == 5
                atom_stereo = SquarePlanar(atoms=sp_atoms, parity=0)

            elif chiral_tag == Chem.ChiralType.CHI_TRIGONALBIPYRAMIDAL:
                perm = atom.GetUnsignedProp("_chiralPermutation")
                tbp_order = self._tbp_atom_order_permutation_dict[perm]
                neigh_atoms = tuple([neighbors[i] for i in tbp_order])
                tbp_atoms = (id_atom_map[atom_idx], *neigh_atoms)
                assert len(tbp_atoms) == 6
                atom_stereo = TrigonalBipyramidal(tbp_atoms, 1)

            elif chiral_tag == Chem.ChiralType.CHI_OCTAHEDRAL:
                perm = atom.GetUnsignedProp("_chiralPermutation")
                order = self._oct_atom_order_permutation_dict[perm]
                neigh_atoms = tuple([neighbors[i] for i in order])
                oct_atoms = (id_atom_map[atom_idx], *neigh_atoms)
                assert len(oct_atoms) == 7
                atom_stereo = Octahedral(oct_atoms, 1)

            else:
                continue
            
            if not self.lone_pair_stereo and None in atom_stereo.atoms:
                continue
            graph.set_atom_stereo(atom_stereo)

        for bond in (
            b
            for b in rdmol.GetBonds()
            if (rd_bond_stereo := b.GetStereo()) in self._rd_atrop.keys()
            or b.GetIsAromatic()
            or b.GetBondType() == Chem.rdchem.BondType.DOUBLE
        ):
            bond_stereo: BondStereo

            begin_idx: int = bond.GetBeginAtomIdx()
            end_idx: int = bond.GetEndAtomIdx()

            neighbors_begin: list[int] = [
                atom.GetIdx()
                for atom in rdmol.GetAtomWithIdx(begin_idx).GetNeighbors()
                if atom.GetIdx() != end_idx
            ]

            neighbors_end: list[int] = [
                atom.GetIdx()
                for atom in rdmol.GetAtomWithIdx(end_idx).GetNeighbors()
                if atom.GetIdx() != begin_idx
            ]
            if len(neighbors_begin) > 2 or len(neighbors_end) > 2:
                continue

            if (
                rd_bond_stereo == Chem.BondStereo.STEREOATROPCW
                or rd_bond_stereo == Chem.BondStereo.STEREOATROPCCW
            ):
                atrop_parity = self._rd_atrop[rd_bond_stereo]
                stereo_atoms = [a for a in bond.GetStereoAtoms()]

                if (
                    stereo_atoms[0] in neighbors_begin
                    and stereo_atoms[1] in neighbors_end
                ):
                    bond_atoms_idx = (
                        stereo_atoms[0],
                        *[n for n in neighbors_begin if n != stereo_atoms[0]],
                        begin_idx,
                        end_idx,
                        stereo_atoms[1],
                        *[n for n in neighbors_end if n != stereo_atoms[1]],
                    )

                    bond_atoms = tuple(
                        [id_atom_map[a] for a in bond_atoms_idx]
                    )

                elif (
                    stereo_atoms[0] in neighbors_end
                    and stereo_atoms[1] in neighbors_begin
                ):
                    bond_atoms_idx = (
                        stereo_atoms[0],
                        *[n for n in neighbors_end if n != stereo_atoms[0]],
                        begin_idx,
                        end_idx,
                        stereo_atoms[1],
                        *[n for n in neighbors_begin if n != stereo_atoms[1]],
                    )

                    bond_atoms = tuple(
                        [id_atom_map[a] for a in bond_atoms_idx]
                    )
                else:
                    raise RuntimeError("Stereo Atoms not neighbors")

                assert len(bond_atoms) == 6
                bond_stereo = AtropBond(bond_atoms, atrop_parity)

            elif rd_bond_stereo in (
                Chem.BondStereo.STEREOZ,
                Chem.BondStereo.STEREOE,
            ):
                invert = {
                    Chem.BondStereo.STEREOZ: False,
                    Chem.BondStereo.STEREOE: True,
                }[rd_bond_stereo]

                begin_stereo_atom: int
                end_stereo_atom: int
                begin_stereo_atom, end_stereo_atom = [
                    a for a in bond.GetStereoAtoms()
                ]
                begin_non_stereo_nbr = (
                    None
                    if len(neighbors_begin) == 1
                    else [
                        a for a in neighbors_begin if a != begin_stereo_atom
                    ][0]
                )
                end_non_stereo_nbr = (
                    None
                    if len(neighbors_end) == 1
                    else [a for a in neighbors_end if a != end_stereo_atom][0]
                )

                bond_atoms_idx = (
                    begin_stereo_atom,
                    begin_non_stereo_nbr,
                    begin_idx,
                    end_idx,
                    end_stereo_atom,
                    end_non_stereo_nbr,
                )
                assert len(bond_atoms_idx) == 6

                assert len(bond_atoms_idx) == 6

                bond_atoms = tuple(
                    [id_atom_map.get(a) for a in bond_atoms_idx]
                )
 
                if invert:
                    bond_atoms = tuple(
                        [bond_atoms[i] for i in (1, 0, 2, 3, 4, 5)]
                        )
                assert len(bond_atoms) == 6
                bond_stereo = PlanarBond(bond_atoms, 0)
                
            elif bond.GetBondType() == Chem.rdchem.BondType.AROMATIC or (
                bond.GetBondType() == Chem.rdchem.BondType.DOUBLE
                and rd_bond_stereo == Chem.BondStereo.STEREONONE
            ):
                # Find rings with bond begin_idx-end_idx, sort by aromatic first then size
                rings = [
                    (
                        
                        all(
                            rdmol.GetBondBetweenAtoms(
                                list(r)[i], list(r)[(i + 1) % len(r)]
                            ).GetIsAromatic()
                            for i in range(len(r))
                        ),
                        len(r),
                        list(r),
                    )
                    for r in Chem.GetSymmSSSR(rdmol)
                    if begin_idx in r
                    and end_idx in r
                ]
                rings.sort(key=lambda x: (x[0], x[1]), reverse=True)

                if rings and (rings[0][0] is True # aromatic rings always cis
                              or rings[0][1] < self._min_trans_ring_size):
                    ring = rings[0][2] if rings else []

                    # Get ordered atoms from first ring (assumed cis)

                    n_begin = [
                        n.GetIdx()
                        for n in rdmol.GetAtomWithIdx(begin_idx).GetNeighbors()
                        if n.GetIdx() != end_idx
                     ]
                    n_end = [
                        n.GetIdx()
                        for n in rdmol.GetAtomWithIdx(end_idx).GetNeighbors()
                        if n.GetIdx() != begin_idx
                    ]

                    bond_atoms_idx = [
                        next(
                        (n for n in n_begin if n in ring), None
                         ),  # atom1: in-ring neighbor of begin_idx
                        next(
                        (n for n in n_begin if n not in ring), None
                        ),  # atom2: out-of-ring neighbor of begin_idx
                        begin_idx,  # atom3
                        end_idx,  # atom4
                        next(
                        (n for n in n_end if n in ring), None
                        ),  # atom5: in-ring neighbor of end_idx
                        next(
                        (n for n in n_end if n not in ring), None
                        ),  # atom6: out-of-ring neighbor of end_idx
                    ]
                    bond_atoms = tuple(
                        [id_atom_map.get(a) for a in bond_atoms_idx]
                    )

                    assert len(bond_atoms) == 6
                    
                    
                    bond_stereo = PlanarBond(bond_atoms, 0)
                        
                else:
                    neighbors_begin_with_none = neighbors_begin + [None] * (2 - len(neighbors_begin))
                    neighbors_end_with_none = neighbors_end + [None] * (2 - len(neighbors_end))
                    bond_atoms_idx = (
                    *neighbors_begin_with_none,

                    begin_idx,
                    end_idx,
                    *neighbors_end_with_none,
                    )

                    bond_atoms = tuple(
                    [id_atom_map.get(a) for a in bond_atoms_idx]
                    )
                    assert len(bond_atoms) == 6, bond_atoms

                    bond_stereo = PlanarBond(bond_atoms, None)
                    bond_stereo = PlanarBond(bond_atoms, None)
                    
            else:
                continue
        
            if not self.lone_pair_stereo and None in bond_stereo.atoms:
                continue
            elif (isinstance(bond_stereo, (PlanarBond, AtropBond))):
                if all(bond_stereo.atoms[a] is None for a in (0, 1)):
                    continue
                elif all(bond_stereo.atoms[a] is None for a in (4, 5)):
                    continue
            graph.set_bond_stereo(bond_stereo)

        return graph

    _rd_tetrahedral: ClassVar[
        Mapping[Chem.rdchem.ChiralType, None | Literal[-1, 1]]
    ] = MappingProxyType(
        {
            Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW: -1,
            Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW: 1,
            Chem.rdchem.ChiralType.CHI_TETRAHEDRAL: None,
        }
    )

    _rd_atrop: ClassVar[Mapping[Chem.BondStereo, None | Literal[-1, 1]]] = (
        MappingProxyType(
            {
                Chem.BondStereo.STEREOATROPCW: 1,
                Chem.BondStereo.STEREOATROPCCW: -1,
            }
        )
    )

    _tbp_atom_order_permutation_dict = MappingProxyType(
        {
            1: (0, 1, 2, 3, 4),
            2: (0, 1, 3, 2, 4),
            3: (0, 1, 2, 4, 3),
            4: (0, 1, 4, 2, 3),
            5: (0, 1, 3, 4, 2),
            6: (0, 1, 4, 3, 2),
            7: (0, 2, 3, 4, 1),
            8: (0, 2, 4, 3, 1),
            9: (1, 0, 2, 3, 4),
            11: (1, 0, 3, 2, 4),
            10: (1, 0, 2, 4, 3),
            12: (1, 0, 4, 2, 3),
            13: (1, 0, 3, 4, 2),
            14: (1, 0, 4, 3, 2),
            15: (2, 0, 1, 3, 4),
            16: (2, 0, 1, 4, 3),
            17: (3, 0, 1, 2, 4),
            18: (3, 0, 2, 1, 4),
            19: (2, 0, 4, 1, 3),
            20: (2, 0, 3, 1, 4),
        }
    )
    "adapted from http://opensmiles.org/opensmiles.html"

    _oct_atom_order_permutation_dict = MappingProxyType(
        {
            1: (0, 5, 1, 2, 3, 4),
            2: (0, 5, 1, 4, 3, 2),
            3: (0, 4, 1, 2, 3, 5),
            16: (0, 4, 1, 5, 3, 2),
            6: (0, 3, 1, 2, 4, 5),
            18: (0, 3, 1, 5, 4, 2),
            19: (0, 2, 1, 3, 4, 5),
            24: (0, 2, 1, 5, 4, 3),
            25: (0, 1, 2, 3, 4, 5),
            30: (0, 1, 2, 5, 4, 3),
            4: (0, 5, 1, 2, 4, 3),
            14: (0, 5, 1, 3, 4, 2),
            5: (0, 4, 1, 2, 5, 3),
            15: (0, 4, 1, 3, 5, 2),
            7: (0, 3, 1, 2, 5, 4),
            17: (0, 3, 1, 4, 5, 2),
            20: (0, 2, 1, 3, 5, 4),
            23: (0, 2, 1, 4, 5, 3),
            26: (0, 1, 2, 3, 5, 4),
            29: (0, 1, 2, 4, 5, 3),
            10: (0, 5, 1, 4, 2, 3),
            8: (0, 5, 1, 3, 2, 4),
            11: (0, 4, 1, 5, 2, 3),
            9: (0, 4, 1, 3, 2, 5),
            13: (0, 3, 1, 5, 2, 4),
            12: (0, 3, 1, 4, 2, 5),
            22: (0, 2, 1, 5, 3, 4),
            21: (0, 2, 1, 4, 3, 5),
            28: (0, 1, 2, 5, 3, 4),
            27: (0, 1, 2, 4, 3, 5),
        }
    )



from __future__ import annotations

import itertools
import sys
from collections import Counter
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    Protocol,
    TypeVar,
    runtime_checkable,
)

if sys.version_info >= (3, 13):
    from typing import TypeVar
else:
    from typing_extensions import TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Set

    from typing_extensions import Self

    from stereomolgraph.graphs.mg import AtomId, Bond


OInt = None | int
"Optional Integer"

A = TypeVar(
    "A", bound=tuple[OInt, ...],
    covariant=True,
    default=tuple[OInt, ...]
)
P = TypeVar(
    "P",
    covariant=True,
    bound=None | Literal[1, 0, -1],
    default=None | Literal[1, 0, -1],
)


@runtime_checkable
class Stereo(Protocol, Generic[A, P]):
    """
    Protocol to represent the orientation of a group of atoms in space.
    This is used to represent local stereochemistry and simultaneously the
    hybridization of atoms.
    """

    atoms: A
    """Atoms are a order dependent tuple of integers."""

    parity: P
    """parity is a number that defines the orientation of the atoms. If None,
        the relative orientation of the atoms is not defined.
        If 0 the orientation is defined and part of a achiral stereochemistry.
        If 1 or -1 the orientation is defined and part of a chiral stereochemistry.
        """
    @property
    def PERMUTATION_GROUP(self,) -> Iterable[A]:
        """Defines all allowed permutations defined by the symmetry group under
        which the stereochemistry is invariant."""
        ...

    def __init__(self, atoms: A, parity: P = None): ...

    def __eq__(self, other: Any) -> bool: ...

    def __hash__(self) -> int: ...

    def invert(self) -> Self:
        """Inverts the stereo. If the stereo is achiral, it returns itself."""
        ...

    def get_isomers(self) -> Set[Self]:
        """Returns all stereoisomers of the stereochemistry. Not just the
        inverted ones, but all possible stereoisomers."""
        ...


@runtime_checkable
class AtomStereo(Stereo[A, P], Protocol, Generic[A, P]):
    @property
    def central_atom(self) -> AtomId: ...


@runtime_checkable
class BondStereo(Stereo[A, P], Protocol, Generic[A, P]):
    @property
    def bond(self) -> Bond: ...


class _StereoMixin(Generic[A, P]):
    PERMUTATION_GROUP: Iterable[A]
    inversion: None | A
    atoms: A
    parity: P

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.atoms}, {self.parity})"

    def __init__(self, atoms: A, parity: P = None):
        self.atoms = atoms
        self.parity = parity

    def _perm_atoms(self) -> Iterator[A]:
        if self.parity is None:
            return (
                tuple([self.atoms[i] for i in perm])
                for perm in itertools.permutations(range(len(self.atoms)))
            )
        else:
            return (
                tuple([self.atoms[i] for i in perm])
                for perm in self.PERMUTATION_GROUP
            )

    def invert(self) -> Self:
        if self.parity is None:
            return self
        if self.parity == 0:
            return self
        new_parity = -self.parity
        assert new_parity in (1, -1)
        return self.__class__(self.atoms, new_parity) # type: ignore[return-value]

    def _inverted_atoms(self) -> A:
        if self.inversion is None:
            return self.atoms
        atoms = tuple([self.atoms[i] for i in self.inversion])
        assert len(atoms) == len(self.atoms) == len(self.inversion)
        return atoms  # type: ignore[return-value]

    def __eq__(self, other: Any) -> bool:
        if not hasattr(other, "atoms") or not hasattr(other, "parity"):
            return NotImplemented
        s_atoms, o_atoms = self.atoms, other.atoms
        set_s_atoms = set(s_atoms)
        set_o_atoms = set(o_atoms)

        if self.parity is None or other.parity is None:
            if set_s_atoms == set_o_atoms:
                return True
            return False

        if self.parity in (1, -1):
            if other.parity == 0:
                return False

            if len(s_atoms) != len(o_atoms) or not set_s_atoms.issuperset(
                set_o_atoms
            ):
                return False

            elif self.parity == other.parity:
                return o_atoms == s_atoms or any(
                    o_atoms == p for p in self._perm_atoms()
                )

            elif self.parity * -1 == other.parity:
                return any(
                    other._inverted_atoms() == p for p in self._perm_atoms()
                )

        if self.parity == 0:
            if other.parity in (1, -1):
                return False

            if len(s_atoms) != len(o_atoms) or not set_s_atoms.issuperset(
                set_o_atoms
            ):
                return False

            if self.parity is None or other.parity is None:
                return set_s_atoms == set_o_atoms

            if self.parity == other.parity:
                return o_atoms == s_atoms or o_atoms in self._perm_atoms()
        raise RuntimeError(
            "This should not happen! "
            "Either the parity is not set correctly or the atoms are not "
            "ordered correctly."
        )

    def __hash__(self) -> int:
        if self.parity is None:
            return hash(frozenset(Counter(self.atoms).items()))
        elif self.parity == 0:
            perm = frozenset(
                {
                    tuple([self.atoms[i] for i in perm])
                    for perm in self.PERMUTATION_GROUP
                }
            )
            return hash(perm)
        # else parity in (1, -1):
        perm = frozenset(
            {
                tuple([self.atoms[i] for i in perm])
                for perm in self.PERMUTATION_GROUP
            }
        )

        inverted_perm = frozenset(
            {
                tuple([self._inverted_atoms()[i] for i in perm])
                for perm in self.PERMUTATION_GROUP
            }
        )

        if self.parity == 1:
            return hash((perm, inverted_perm))
        elif self.parity == -1:
            return hash((inverted_perm, perm))
        else:
            raise RuntimeError("Something is wrong with parity")


class Tetrahedral(
    _StereoMixin[tuple[OInt, OInt, OInt, OInt, OInt], None | Literal[1, -1]],
):
    r"""Represents all possible configurations of atoms for a Tetrahedral
    Stereochemistry::

       parity = 1      parity = -1
           4                4
           |                |
           0                0
        /  ¦  \          /  ¦  \
       2   1   3        3   1   2

    Atoms of the tetrahedral stereochemistry are ordered in a way that when the
    first atom is rotated to the back, the other atoms in order are rotated in
    the direction defined by the stereo.

    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    :ivar PERMUTATION_GROUP: Permutations allowed by the stereochemistry
    """

    inversion = (0, 2, 1, 3, 4)
    PERMUTATION_GROUP = (
            (0, 1, 2, 3, 4),
            (0, 3, 1, 2, 4),
            (0, 2, 3, 1, 4),
            (0, 1, 4, 2, 3),
            (0, 2, 1, 4, 3),
            (0, 4, 2, 1, 3),
            (0, 1, 3, 4, 2),
            (0, 4, 1, 3, 2),
            (0, 3, 4, 1, 2),
            (0, 2, 4, 3, 1),
            (0, 3, 2, 4, 1),
            (0, 4, 3, 2, 1),
    )

    def get_isomers(self) -> set[Self]:
        return {
            self.__class__(atoms=self.atoms, parity=1),
            self.__class__(atoms=self.atoms, parity=-1),
        }

    @property
    def central_atom(self) -> AtomId:
        return self.atoms[0]


class SquarePlanar(
    _StereoMixin[tuple[OInt, OInt, OInt, OInt, OInt], None | Literal[0]],
):
    r""" Represents all possible configurations of atoms for a
    SquarePlanar Stereochemistry::

        1     4
         \   /
           0
         /   \
        2     3

    Atoms of the Square Planar stereochemistry are ordered in a way that


    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    """

    inversion = None
    PERMUTATION_GROUP = (
            (0, 1, 2, 3, 4),
            (0, 2, 3, 4, 1),
            (0, 3, 4, 1, 2),
            (0, 4, 1, 2, 3),
            (0, 4, 3, 2, 1),
            (0, 3, 2, 1, 4),
            (0, 2, 1, 4, 3),
            (0, 1, 4, 3, 2),
    )

    def get_isomers(self) -> set[SquarePlanar]:
        return {
            SquarePlanar(atoms=atoms, parity=0)
            for perm in itertools.permutations(self.atoms[1:])
            if len(atoms := (self.atoms[0], *perm)) == 5
        }

    @property
    def central_atom(self) -> AtomId:
        return self.atoms[0]


class TrigonalBipyramidal(
    _StereoMixin[tuple[OInt, OInt, OInt, OInt, OInt, OInt], None | Literal[1, -1]],
):
    r"""Represents all possible configurations of atoms for a
    TrigonalBipyramidal Stereochemistry::

       parity = 1             parity = -1
        3   1                     1   3
         ◁  ¦                    ¦  ▷
            0  — 5           5 —  0
         ◀  ¦                    ¦  ▶
        4   2                     2   4

    Atoms of the trigonal bipyramidal stereochemistry are ordered in a way that
    when the first two atoms are the top and bottom of the bipyramid. The last
    three equatorial atoms are ordered in a way that when the first atom is
    rotated to the back, the other atoms in order are rotated in the direction
    defined by the stereo.

    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    """

    inversion = (0, 1, 2, 3, 5, 4)
    PERMUTATION_GROUP = (
            (0, 1, 2, 3, 4, 5),
            (0, 1, 2, 5, 3, 4),
            (0, 1, 2, 4, 5, 3),
            (0, 2, 1, 3, 5, 4),
            (0, 2, 1, 5, 4, 3),
            (0, 2, 1, 4, 3, 5),
    )

    def get_isomers(self) -> set[Self]:
        return {
            self.__class__(atoms=atoms, parity=p)
            for perm in itertools.permutations(self.atoms[1:])
            for p in (1, -1)
            if len(atoms := (self.atoms[0], *perm)) == 6
        }

    @property
    def central_atom(self) -> AtomId:
        return self.atoms[0]


class Octahedral(
    _StereoMixin[
        tuple[OInt, OInt, OInt, OInt, OInt, OInt, OInt], None | Literal[1, -1]
    ],
):
    """Represents all possible configurations of atoms for a Octahedral
    Stereochemistry::

        parity = 1             parity = -1
         3  1   6                3  2  6
          ◁ ¦ /                  ◁ ¦ /
            0                       0
          / ¦ ▶                  / ¦  ▶
         4  2  5                4   1  5
    """

    inversion = (0, 2, 1, 3, 4, 5, 6)
    PERMUTATION_GROUP = (
            (0, 1, 2, 3, 4, 5, 6),
            (0, 1, 2, 6, 3, 4, 5),
            (0, 1, 2, 5, 6, 3, 4),
            (0, 1, 2, 4, 5, 6, 3),
            (0, 2, 1, 4, 3, 6, 5),
            (0, 2, 1, 5, 4, 3, 6),
            (0, 2, 1, 6, 5, 4, 3),
            (0, 2, 1, 3, 6, 5, 4),
            (0, 3, 5, 2, 4, 1, 6),
            (0, 3, 5, 6, 2, 4, 1),
            (0, 3, 5, 1, 6, 2, 4),
            (0, 3, 5, 4, 1, 6, 2),
            (0, 5, 3, 1, 4, 2, 6),
            (0, 5, 3, 6, 1, 4, 2),
            (0, 5, 3, 2, 6, 1, 4),
            (0, 5, 3, 4, 2, 6, 1),
            (0, 4, 6, 3, 2, 5, 1),
            (0, 4, 6, 1, 3, 2, 5),
            (0, 4, 6, 5, 1, 3, 2),
            (0, 4, 6, 2, 5, 1, 3),
            (0, 6, 4, 3, 1, 5, 2),
            (0, 6, 4, 2, 3, 1, 5),
            (0, 6, 4, 5, 2, 3, 1),
            (0, 6, 4, 1, 5, 2, 3),
    )

    def get_isomers(self) -> set[Octahedral]:
        return {
            Octahedral(atoms=atoms, parity=p)
            for perm in itertools.permutations(self.atoms[1:])
            for p in (1, -1)
            if len((atoms := (self.atoms[0], *perm))) == 7
        }

    @property
    def central_atom(self) -> AtomId:
        return self.atoms[0]


class PlanarBond(
    _StereoMixin[tuple[OInt, OInt, OInt, OInt, OInt, OInt], None | Literal[0]],
):
    r""" Represents all possible configurations of atoms for a
    Planar Structure and should be used for aromatic and double bonds::

        0        4
         \      /
          2 == 3
         /      \
        1        5

    All atoms of the double bond are in one plane. Atoms 2 and 3 are the center
    Atoms 0 and 1 are bonded to 2 and atoms 4 and 5 are bonded to 3.
    The stereochemistry is defined by the relative orientation
    of the atoms 0, 1, 4 and 5.

    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    :ivar PERMUTATION_GROUP: Permutations allowed by the stereochemistry
    """

    inversion = None
    PERMUTATION_GROUP = (
            (0, 1, 2, 3, 4, 5),
            (1, 0, 2, 3, 5, 4),
            (4, 5, 3, 2, 0, 1),
            (5, 4, 3, 2, 1, 0),
    )

    def get_isomers(self) -> set[PlanarBond]:
        reordered_atoms = tuple(self.atoms[i] for i in (0, 1, 2, 3, 5, 4))
        assert len(reordered_atoms) == 6
        return {
            PlanarBond(self.atoms, 0),
            PlanarBond(reordered_atoms, 0),
        }

    @property
    def bond(self) -> Bond:
        bond = frozenset(self.atoms[2:4])
        assert len(bond) == 2
        return bond


class AtropBond(
    _StereoMixin[tuple[OInt, OInt, OInt, OInt, OInt, OInt], None | Literal[1, -1]],
):
    r"""
    Represents all possible configurations of atoms for a
    Atropostereoisomer bond::
    
        parity = 1          parity = -1
        1       5           1        5
         \     /            ◀      /
          2 - 3               2 - 3
        ◀      \            /      \
        0        4         0         4


    """

    inversion = (1, 0, 2, 3, 4, 5)
    PERMUTATION_GROUP = (
            (0, 1, 2, 3, 4, 5),
            (1, 0, 2, 3, 5, 4),
            (4, 5, 3, 2, 1, 0),
            (5, 4, 3, 2, 0, 1),
    )

    def get_isomers(self) -> set[AtropBond]:
        other_atoms = tuple(self.atoms[i] for i in (0, 1, 2, 3, 5, 4))
        assert len(other_atoms) == 6
        return {
            AtropBond(self.atoms, 1),
            AtropBond(other_atoms, -1),
        }

    @property
    def bond(self) -> Bond:
        bond = frozenset(self.atoms[2:4])
        assert len(bond) == 2
        return bond



from __future__ import annotations

from collections import Counter, defaultdict
from typing import TYPE_CHECKING, NamedTuple

from stereomolgraph.algorithms.color_refine import label_hash

if TYPE_CHECKING:
    from collections.abc import Callable, Hashable, Iterable, Iterator, Mapping
    from typing import TypeVar

    import numpy as np

    from stereomolgraph.graphs import (
        AtomId,
        CondensedReactionGraph,
        MolGraph,
        StereoCondensedReactionGraph,
        StereoMolGraph,
    )
    from stereomolgraph.graphs.scrg import Change
    from stereomolgraph.stereodescriptors import Stereo

    KT = TypeVar("KT", bound=Hashable, covariant=True)
    VT = TypeVar("VT", bound=Hashable, covariant=True)

class _Parameters(NamedTuple):
    """
    Parameters of the algorithm.
    :ivar g1_nbrhd: Neighborhood of the first graph
    :ivar g2_nbrhd: Neighborhood of the second graph
    :ivar g1_labels: Labels of the first graph
    :ivar g2_labels: Labels of the second graph
    :ivar nodes_of_g1Labels: Nodes of the first graph grouped by labels
    :ivar nodes_of_g2Labels: Nodes of the second graph grouped by labels
    :ivar g1_degree: Degree of the first graph
    :ivar g2_nodes_of_degree: Nodes of the second graph grouped by degree
    :ivar g1_stereo: Stereochemistry of the first graph
    :ivar g2_stereo: Stereochemistry of the second graph
    :ivar g1_stereo_changes: Stereochemistry changes of the first graph
    :ivar g2_stereo_changes: Stereochemistry changes of the second graph
    """

    # Neighborhood
    g1_nbrhd: Mapping[AtomId, set[AtomId]]
    g2_nbrhd: Mapping[AtomId, set[AtomId]]
    # atomid: label
    g1_labels: Mapping[AtomId, np.int64]
    g2_labels: Mapping[AtomId, np.int64]
    # label: set of atomids
    nodes_of_g1Labels: Mapping[np.int64, set[AtomId]]
    nodes_of_g2Labels: Mapping[np.int64, set[AtomId]]
    # degree: set of atomids
    g1_degree: Mapping[AtomId, int]
    g2_nodes_of_degree: Mapping[int, set[AtomId]]
    # atomid: list of stereos containing the atom
    g1_stereo: Mapping[AtomId, list[Stereo]]
    g2_stereo: Mapping[AtomId, list[Stereo]]

    g1_stereo_changes: Mapping[AtomId, Mapping[Change, list[Stereo]]]
    g2_stereo_changes: Mapping[AtomId, Mapping[Change, list[Stereo]]]


class _State(NamedTuple):
    """
    State of the algorithm.

    :ivar mapping:
    :ivar inverted_mapping:
    :ivar frontier1: neighbors of mapped atoms in g1
    :ivar external1: atoms not in mapping and not in frontier1
    :ivar frontier2: neighbors of mapped atoms in g2
    :ivar external2: atoms not in mapping and not in frontier2
    """

    mapping: dict[AtomId, AtomId]
    inverted_mapping: dict[AtomId, AtomId]

    frontier1: set[AtomId]
    "neighbors of mapped atoms in g1"
    external1: set[AtomId]
    "atoms not in mapping and not in frontier1"

    frontier2: set[AtomId]
    "neighbors of mapped atoms in g2"
    external2: set[AtomId]
    "atoms not in mapping and not in frontier2"


def _group_keys_by_value(many_to_one: Mapping[KT, VT]) -> dict[VT, set[KT]]:
    """Inverts a many-to-one mapping to create a one-to-many mapping.

    Converts a dictionary where multiple keys may point to the same value
    into a dictionary where each original value maps to a set of all original
    keys.

    >>> _group_keys_by_value({"a": 1, "b": 1, "c": 2, "d": 3, "e": 3})
    {1: {'a', 'b'}, 2: {'c'}, 3: {'e', 'd'}}

    :param many_to_one: Dictionary to invert
    :return: Inverted dictionary where each value maps to a set of keys
    """
    inverted: defaultdict[VT, set[KT]] = defaultdict(set)
    for key, value in many_to_one.items():
        inverted[value].add(key)
    return dict(inverted)


def _bfs_layers(
    neighbor_dict: Mapping[AtomId, Iterable[AtomId]],
    sources: Iterable[AtomId] | AtomId,
) -> Iterator[list[AtomId]]:
    """
    Generates layers of the graph starting from the source atoms.
    Each layer contains all nodes that are at the same distance from the
    sources.
    The first layer contains the sources.

    :param neighbor_dict: Dictionary of neighbors for each atom
    :param sources: Sources to start from
    """
    if isinstance(sources, int):
        sources = [sources]
    
    current_layer = list(sources)
    visited = set(sources)

    if any(source not in neighbor_dict for source in current_layer):
        raise ValueError("Source atom not in molecule")

    # this is basically BFS, except that the current layer only stores the
    # nodes at same distance from sources at each iteration
    while current_layer:
        yield current_layer
        next_layer: list[AtomId] = []
        for node in current_layer:
            for child in neighbor_dict[node]:
                if child not in visited:
                    visited.add(child)
                    next_layer.append(child)
        current_layer = next_layer


def _sanity_check_and_init(
    g1: StereoMolGraph | MolGraph,
    g2: StereoMolGraph | MolGraph,
    atom_labels: None| tuple[np.ndarray[tuple[int], np.dtype[np.int64]],
                            np.ndarray[tuple[int], np.dtype[np.int64]]] = None,
    stereo: bool = False,
    stereo_change: bool = False,
    subgraph: bool = False,
) -> None | tuple[_Parameters, _State]:
    if stereo_change and not stereo:
        raise ValueError("Stereo change is only available for stereo graphs.")

    g1_nbrhd = {a: g1.neighbors[a] for a in g1.atoms}
    g2_nbrhd = {a: g2.neighbors[a] for a in g2.atoms}

    if not subgraph:
        if len(g1_nbrhd) != len(g2_nbrhd):
            return None

        if sorted(len(n) for n in g1_nbrhd.values()) != sorted(
            len(n) for n in g2_nbrhd.values()
        ):
            return None

    elif subgraph:
        if len(g1_nbrhd) > len(g2_nbrhd):
            ValueError("The second graph must be larger than the first one.")
        elif len(g1_nbrhd) == len(g2_nbrhd):
            raise ValueError(
                "Both graphs have the same number of atoms. "
                "Do not use subgraph isomorphism in this case."
            )
        elif Counter(len(nbr) for nbr in g1_nbrhd.values()) > Counter(
            len(nbr) for nbr in g2_nbrhd.values()
        ):
            return None

    if atom_labels is None:
        g1_labels = {a:h for a, h in zip(g1.atoms, label_hash(g1))}
        g2_labels = {a:h for a, h in zip(g2.atoms, label_hash(g2))}
    else:
        g1_labels, g2_labels = atom_labels

    g1_labels_counter = Counter(g1_labels.values())
    g2_labels_counter = Counter(g2_labels.values())

    if not subgraph and g1_labels_counter != g2_labels_counter:
        return None

    elif subgraph and g1_labels_counter > g2_labels_counter:
        return None

    g1_stereo: defaultdict[AtomId, list[Stereo]] = defaultdict(list)
    g2_stereo: defaultdict[AtomId, list[Stereo]] = defaultdict(list)

    if stereo:
        if TYPE_CHECKING:
            assert isinstance(g1, StereoMolGraph)
            assert isinstance(g2, StereoMolGraph)

        for s in g1.stereo.values():
            for atom in s.atoms:
                g1_stereo[atom].append(s)

        for s in g2.stereo.values():
            for atom in s.atoms:
                g2_stereo[atom].append(s)

    g1_stereo_changes: defaultdict[
        AtomId, defaultdict[Change, list[Stereo]]
    ] = defaultdict(lambda: defaultdict(list))

    g2_stereo_changes: defaultdict[
        AtomId, defaultdict[Change, list[Stereo]]
    ] = defaultdict(lambda: defaultdict(list))

    if stereo_change:
        if TYPE_CHECKING:
            assert isinstance(g1, StereoCondensedReactionGraph)
            assert isinstance(g2, StereoCondensedReactionGraph)
        
        for _, stereo_change_dict in g1.atom_stereo_changes.items():
            for stereo_change_enum, atom_stereo in stereo_change_dict.items():
                if atom_stereo is not None:
                    for atom in atom_stereo.atoms:
                        g1_stereo_changes[atom][stereo_change_enum].append(
                            atom_stereo
                        )

        for _, stereo_change_dict in g2.atom_stereo_changes.items():
            for stereo_change_enum, atom_stereo in stereo_change_dict.items():
                if atom_stereo is not None:
                    for atom in atom_stereo.atoms:
                        g2_stereo_changes[atom][stereo_change_enum].append(
                            atom_stereo
                        )

        for _, stereo_change_dict in g1.bond_stereo_changes.items():
            for stereo_change_enum, bond_stereo in stereo_change_dict.items():
                if bond_stereo is not None:
                    for atom in bond_stereo.atoms:
                        g1_stereo_changes[atom][stereo_change_enum].append(
                            bond_stereo
                        )

        for _, stereo_change_dict in g2.bond_stereo_changes.items():
            for stereo_change_enum, bond_stereo in stereo_change_dict.items():
                if bond_stereo is not None:
                    for atom in bond_stereo.atoms:
                        g2_stereo_changes[atom][stereo_change_enum].append(
                            bond_stereo
                        )

    g1_degree = {a: len(n_set) for a, n_set in g1_nbrhd.items()}
    g2_degree = {a: len(n_set) for a, n_set in g2_nbrhd.items()}

    params = _Parameters(
        g1_nbrhd,
        g2_nbrhd,
        g1_labels,
        g2_labels,
        _group_keys_by_value(g1_labels),
        _group_keys_by_value(g2_labels),
        g1_degree,
        _group_keys_by_value(g2_degree),
        g1_stereo,
        g2_stereo,
        g1_stereo_changes,
        g2_stereo_changes,
    )

    state = _State({}, {}, set(), set(g1_nbrhd), set(), set(g2_nbrhd))

    return params, state


def _wrap_all(
    *funcs: Callable[[AtomId, AtomId, _State, _Parameters], bool],
) -> Callable[[AtomId, AtomId, _State, _Parameters], bool]:
    def wrapper(
        a: AtomId, b: AtomId, state: _State, params: _Parameters
    ) -> bool:
        return all(f(a, b, state, params) for f in funcs)

    return wrapper


def vf2pp_all_isomorphisms(
    g1: MolGraph
    | StereoMolGraph
    | CondensedReactionGraph
    | StereoCondensedReactionGraph,
    g2: MolGraph
    | StereoMolGraph
    | CondensedReactionGraph
    | StereoCondensedReactionGraph,
    atom_labels: None| tuple[Mapping[AtomId, int],
                             Mapping[AtomId, int]] = None,
    stereo: bool = False,
    stereo_change: bool = False,
    subgraph: bool = False,
) -> Iterator[dict[AtomId, AtomId]]:
    
    """Find all isomorphisms between two graphs.

    Algorithms are based of VF2++.
    [VF2++ is a fast algorithm for subgraph isomorphism](https://doi.org/10.1016/j.dam.2018.02.018)"""

    if params_state := _sanity_check_and_init(
        g1, g2, atom_labels, stereo, stereo_change, subgraph
    ):
        params, state = params_state
    else:
        return  # if no isomorphisms return like an empty generator

    # setup helper function based on input parameters

    feasibility_funcs: list[
        Callable[[AtomId, AtomId, _State, _Parameters], bool]
    ] = []
    if subgraph:
        feasibility_funcs.append(_subgraph_feasibility)
        if stereo:
            feasibility_funcs.append(_subgraph_stereo_feasibility)
        if stereo_change:
            feasibility_funcs.append(_subgraph_stereo_change_feasibility)
            
    elif not subgraph:
        feasibility_funcs.append(_graph_feasibility)
        if stereo:
            feasibility_funcs.append(_stereo_feasibility)
        if stereo_change:
            feasibility_funcs.append(_stereo_change_feasibility)
    else:
        raise ValueError("Invalid combination of parameters.")

    feasibility = _wrap_all(*feasibility_funcs) # type: ignore
    revert_state = _revert_state
    update_state = _update_state
    find_candidates = (_find_subgraph_candidates if subgraph
                       else _find_candidates)

    # to avoid overhead
    mapping = state.mapping
    inverted_mapping = state.inverted_mapping
    termination_length = len(g1)

    # Initialize the stack
    node_order: list[AtomId] = _matching_order(params)
    candidates: set[AtomId] = find_candidates(node_order[0], state, params)

    stack: list[tuple[AtomId, set[AtomId]]] = []
    stack.append((node_order[0], candidates))

    # Index of the node from the order, currently being examined
    matching_atom_index = 1

    while stack:
        matching_atom, candidates = stack[-1]
        candidate = candidates.pop() if candidates else None
        if candidate is None:
            # If no remaining candidates, return to a previous state,
            # and follow another branch
            stack.pop()
            matching_atom_index -= 1
            if stack:
                # Pop the previously added u-v pair, and look for
                # a different candidate _v for u
                last_atom1, _ = stack[-1]
                last_atom2 = mapping.pop(last_atom1)
                inverted_mapping.pop(last_atom2)
                revert_state(last_atom1, last_atom2, state, params)
            continue

        mapping[matching_atom] = candidate
        inverted_mapping[candidate] = matching_atom

        if feasibility(matching_atom, candidate, state, params):
            if len(mapping) == termination_length:
                yield mapping.copy()
                mapping.pop(matching_atom)
                inverted_mapping.pop(candidate)
                continue

            update_state(matching_atom, candidate, state, params)
            # Append the next node and its candidates to the stack
            matching_atom = node_order[matching_atom_index]
            candidates = find_candidates(matching_atom, state, params)
            stack.append((matching_atom, candidates))
            matching_atom_index += 1

        else:  # if not feaseble
            mapping.pop(matching_atom)
            inverted_mapping.pop(candidate)


def _graph_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
):
    g1_nbrhd, g2_nbrhd, g1_labels, g2_labels, *_ = params
    _, _, frontier1, external1, frontier2, external2 = state

    t1_labels: list[int] = []
    t2_labels: list[int] = []
    t1_tilde_labels: list[int] = []
    t2_tilde_labels: list[int] = []

    for n in g1_nbrhd[u]:
        if n in external1:
            t1_tilde_labels.append(g1_labels[n])
        elif n in frontier1:
            t1_labels.append(g1_labels[n])

    for n in g2_nbrhd[v]:
        if n in external2:
            t2_tilde_labels.append(g2_labels[n])
        elif n in frontier2:
            t2_labels.append(g2_labels[n])

    t1_labels.sort()
    t2_labels.sort()

    if t1_labels != t2_labels:
        return False

    t1_tilde_labels.sort()
    t2_tilde_labels.sort()

    if t1_tilde_labels != t2_tilde_labels:
        return False

    return True

def _subgraph_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool:
    g1_nbrhd, g2_nbrhd, g1_labels, g2_labels, *_ = params
    _, _, frontier1, external1, frontier2, external2 = state

    counter1 = Counter(g1_labels[n] for n in g1_nbrhd[u] if n in frontier1)
    counter2 = Counter(g2_labels[n] for n in g2_nbrhd[v] if n in frontier2)
    if counter1 > counter2:
        return False

    counter1 = Counter(g1_labels[n] for n in g1_nbrhd[u] if n in external1)
    counter2 = Counter(g2_labels[n] for n in g2_nbrhd[v] if n in external2)
    return counter1 <= counter2

def _stereo_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool:
    s1 = [
        stereo.__class__(
            atoms=tuple([state.mapping[a] for a in stereo.atoms]),
            parity=stereo.parity,
        )
        for stereo in params.g1_stereo[u]
        if all([a in state.mapping for a in stereo.atoms])
    ]

    s2 = [
        stereo
        for stereo in params.g2_stereo[v]
        if all([a in state.inverted_mapping for a in stereo.atoms])
    ]

    if len(s2) != len(s1):
        return False

    if all(s in s2 for s in s1):
        return True
    return False

def _subgraph_stereo_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool: ... # TODO
 
def _stereo_change_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool:
    s1 = {
        (
            stereo_change,
            stereo.__class__(
                atoms=tuple([state.mapping[a] for a in stereo.atoms]),
                parity=stereo.parity,
            ),
        )
        for stereo_change, stereo_list in params.g1_stereo_changes[u].items()
        for stereo in stereo_list
        if stereo is not None # type: ignore
        and all([a in state.mapping for a in stereo.atoms])
    }

    s2 = {
        (stereo_change, stereo)
        for stereo_change, stereo_list in params.g2_stereo_changes[u].items()
        for stereo in stereo_list
        if stereo is not None # type: ignore
        and all([a in state.inverted_mapping for a in stereo.atoms])
    }

    if s1 == s2:
        return True
    return False

def _subgraph_stereo_change_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool: ...
    # TODO: Implement subgraph stereo change feasibility check

def _matching_order(params: _Parameters) -> list[AtomId]:
    g1_nbrhd, _, g1_labels, _, _, nodes_of_g2Labels, *_ = params

    V1_unordered = set(g1_labels.keys())
    label_rarity = {
        label: len(nodes) for label, nodes in nodes_of_g2Labels.items()
    }
    used_degrees = {node: 0 for node in g1_nbrhd}
    node_order: list[int] = []

    while V1_unordered:
        max_rarity = min(label_rarity[g1_labels[x]] for x in V1_unordered)
        rarest_nodes = [
            n for n in V1_unordered if label_rarity[g1_labels[n]] == max_rarity
        ]
        # Use a key that always returns an int (degree) so max() never
        # receives None values. Previously a dict.get could return None
        # for atoms not present in the neighborhood mapping which made
        # `max` try to compare None values and raised a TypeError.
        max_node: AtomId = max(
            rarest_nodes,
            key=lambda a: len(g1_nbrhd.get(a, ())),
        )
        assert isinstance(max_node, int)
        for dlevel_nodes in _bfs_layers(g1_nbrhd, max_node):
            nodes_to_add = dlevel_nodes.copy()
            while nodes_to_add:
                max_used_degree = max(used_degrees[n] for n in nodes_to_add)
                max_used_degree_nodes = [
                    n
                    for n in nodes_to_add
                    if used_degrees[n] == max_used_degree
                ]
                max_degree = max(
                    len(g1_nbrhd[n]) for n in max_used_degree_nodes
                )
                max_degree_nodes = [
                    n
                    for n in max_used_degree_nodes
                    if len(g1_nbrhd[n]) == max_degree
                ]
                next_node = min(
                    max_degree_nodes, key=lambda x: label_rarity[g1_labels[x]]
                )

                node_order.append(next_node)
                for node in g1_nbrhd[next_node]:
                    used_degrees[node] += 1

                nodes_to_add.remove(next_node)
                label_rarity[g1_labels[next_node]] -= 1
                V1_unordered.discard(next_node)

    return node_order


def _find_candidates(
    u: AtomId, state: _State, params: _Parameters
) -> set[AtomId]:
    (
        g1_nbrhd,
        g2_nbrhd,
        g1_labels,
        _,
        _,
        nodes_of_g2Labels,
        g1_deg,
        g2_a_of_deg,
        *_,
    ) = params
    mapping, inverted_mapping, *_, external2 = state

    covered_nbrs = [nbr for nbr in g1_nbrhd[u] if nbr in mapping]

    if not covered_nbrs:
        candidates = set(nodes_of_g2Labels[g1_labels[u]])
        candidates.intersection_update(external2, g2_a_of_deg[g1_deg[u]])
        candidates.difference_update(inverted_mapping)

        return candidates

    nbr1 = covered_nbrs[0]
    candidates = set(g2_nbrhd[mapping[nbr1]])

    for nbr1 in covered_nbrs[1:]:
        candidates.intersection_update(g2_nbrhd[mapping[nbr1]])

    candidates.difference_update(inverted_mapping)
    candidates.intersection_update(
        nodes_of_g2Labels[g1_labels[u]], g2_a_of_deg[g1_deg[u]]
    )
    return candidates


def _find_subgraph_candidates(
    u: AtomId, state: _State, params: _Parameters
) -> set[AtomId]:
    (
        g1_nbrhd,
        g2_nbrhd,
        g1_labels,
        _,
        _,
        nodes_of_g2Labels,
        *_,
    ) = params
    mapping, inverted_mapping, *_, external2 = state

    covered_nbrs = [nbr for nbr in g1_nbrhd[u] if nbr in mapping]

    if not covered_nbrs:
        candidates = set(nodes_of_g2Labels[g1_labels[u]])
        candidates.intersection_update(external2)
        candidates.difference_update(inverted_mapping)
        # candidates.intersection_update(g2_a_of_deg[degree] for degree
        #                               in g2_a_of_deg if degree < g1_deg[u])
        return candidates

    nbr1 = covered_nbrs[0]
    common_nodes = set(g2_nbrhd[mapping[nbr1]])

    for nbr1 in covered_nbrs[1:]:
        common_nodes.intersection_update(g2_nbrhd[mapping[nbr1]])

    common_nodes.difference_update(inverted_mapping)
    common_nodes.intersection_update(nodes_of_g2Labels[g1_labels[u]])

    # common_nodes.intersection_update(g2_a_of_deg[degree] for degree
    #                                 in g2_a_of_deg if degree < g1_deg[u])

    return common_nodes


def _update_state(
    new_atom1: AtomId, new_atom2: AtomId, state: _State, params: _Parameters
) -> None:
    g1_nbrhd, g2_nbrhd, *_ = params
    mapping, inverted_mapping, frontier1, external1, frontier2, external2 = (
        state
    )

    unmapped_neighbors1 = {n for n in g1_nbrhd[new_atom1] if n not in mapping}
    frontier1 |= unmapped_neighbors1
    external1 -= unmapped_neighbors1
    frontier1.discard(new_atom1)
    external1.discard(new_atom1)

    unmapped_neighbors2 = {
        n for n in g2_nbrhd[new_atom2] if n not in inverted_mapping
    }
    frontier2 |= unmapped_neighbors2
    external2 -= unmapped_neighbors2
    frontier2.discard(new_atom2)
    external2.discard(new_atom2)

    return


def _revert_state(
    last_atom1: AtomId, last_atom2: AtomId, state: _State, params: _Parameters
) -> None:
    # If the node we want to remove from the mapping, has at least one
    # covered neighbor, add it to frontier1.
    g1_nbrhd, g2_nbrhd, *_ = params
    mapping, inverted_mapping, frontier1, external1, frontier2, external2 = (
        state
    )

    has_covered_neighbor = False
    for neighbor in g1_nbrhd[last_atom1]:
        if neighbor in mapping:
            # if a neighbor of the excluded node1 is in the mapping,
            # keep node1 in frontier1
            has_covered_neighbor = True
            frontier1.add(last_atom1)
        else:
            # check if its neighbor has another connection with a covered node.
            # If not, only then exclude it from frontier1
            if any(nbr in mapping for nbr in g1_nbrhd[neighbor]):
                continue
            frontier1.discard(neighbor)
            external1.add(neighbor)

    # Case where the node is not present in neither the mapping nor frontier1.
    # By definition, it should belong to external1
    if not has_covered_neighbor:
        external1.add(last_atom1)

    has_covered_neighbor = False
    for neighbor in g2_nbrhd[last_atom2]:
        if neighbor in inverted_mapping:
            has_covered_neighbor = True
            frontier2.add(last_atom2)
        else:
            if any(nbr in inverted_mapping for nbr in g2_nbrhd[neighbor]):
                continue
            frontier2.discard(neighbor)
            external2.add(neighbor)

    if not has_covered_neighbor:
        external2.add(last_atom2)

def stereo_induced_subgraph_mappings(g1: MolGraph, g2:MolGraph,
                                     ) -> Iterator[dict[AtomId, AtomId]]:
    """
    g2 is a induced subgraph of g1 if all atoms of g2 are in g1 and all bonds
    of g2 are in g1 without additional bonds.
    Also the stereodescriptors have to match fully. Additional AtomIds without
    present atoms in g2 will still be mapped. Atom Stereodescriptors have to
    contain the ce
    """
    ...





from __future__ import annotations

import itertools
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Callable, Collection, Iterator
    from typing import Literal, TypeVar

    from stereomolgraph.graphs import (
        CondensedReactionGraph,
        MolGraph,
        StereoCondensedReactionGraph,
        StereoMolGraph,
    )
    from stereomolgraph.stereodescriptors import AtomStereo, BondStereo, Stereo

    S = TypeVar("S", bound=Stereo)
    AS = TypeVar("AS", bound=AtomStereo)
    BS = TypeVar("BS", bound=BondStereo)
    N = TypeVar("N", bound=int)


def numpy_int_tuple_hash(
    arr: np.ndarray[tuple[int, ...], np.dtype[np.int64]],
    out: None | np.ndarray[tuple[Literal[1], ...], np.dtype[np.int64]] = None,
) -> np.ndarray:
    """
    Mimics the python SipHash hashing function for tuples of integers
    with numpy int64 arrays.

    def SipHash(arr_slice):
        h = 0x345678
        mult = 1000003
        length = len(arr_slice)
        for i in range(1, length + 1):
            h = (h ^ arr_slice[i-1]) * mult
            mult += 82520 + 2 * (length - i)
        return h + 97531
    """
    # overflow is an expected behavior in this case
    with np.errstate(over="ignore"):
        arr_shape = arr.shape
        length = arr_shape[-1]
        if out is None:
            output = np.full(arr_shape[:-1], 0x345678, dtype=np.int64)
        else:
            output = out
            output.fill(0x345678)

        n = 82518 + 2 * length
        m = range(n, n - 2 * (length - 1), -2)
        mults = itertools.accumulate(m, initial=1000003)

        for idx, mult in enumerate(mults):
            output ^= arr[..., idx]
            output *= mult

        output += 97531
        return output


def numpy_int_multiset_hash(
    arr: np.ndarray[tuple[int, ...], np.dtype[np.int64]],
    out: None | np.ndarray[tuple[Literal[1], ...], np.dtype[np.int64]] = None,
) -> np.ndarray:
    """
    Hash function for a multiset (order-independent with duplicates) of integers.
    Works by sorting the elements and then applying the tuple hashing function.
    """
    sorted_arr = np.sort(arr, axis=-1)
    return numpy_int_tuple_hash(sorted_arr, out)


def label_hash(
    mg: MolGraph,
    atom_labels: Collection[str] = ("atom_type",),
) -> np.ndarray[tuple[int], np.dtype[np.int64]]:
    """Generates a hash for each atom based on choosen attributes.

    :param mg: MolGraph object containing the atoms.
    :param atom_labels: Iterable of attribute names to use for hashing.
    """
    if atom_labels == ("atom_type",):
        atom_hash = mg.atom_types
    else:
        atom_hash = [
            hash(
                frozenset(
                    (attr, mg.get_atom_attribute(atom, attr))
                    for attr in atom_labels
                )
            )
            for atom in mg.atoms
        ]
    return np.array(atom_hash, dtype=np.int64)


def morgan_generator(
    mg: MolGraph,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> Iterator[np.ndarray[tuple[int], np.dtype[np.int64]]]:
    """Color refinement algorithm for MolGraph.

    This algorithm refines the atom coloring based on their connectivity.
    Identical to the Weisfeiler-Lehman (1-WL) algorithm.

    :param mg: MolGraph object containing the atoms and their connectivity.
    :param max_iter: Maximum number of iterations for refinement.
        Default is None, which means it will run until convergence."""
    n_atoms = len(mg.atoms)
    if atom_labels is not None:
        assert len(atom_labels) == n_atoms

    atom_hash = (
        label_hash(mg, ("atom_type",)) if atom_labels is None else atom_labels
    )
    atom_hash_view = atom_hash.view()
    atom_hash_view.setflags(write=False)
    yield atom_hash_view
    
    if n_atoms == 0:
        return

    arr_id_dict, id_arr_dict = {}, {}
    for id, atom in enumerate(mg.atoms):
        arr_id_dict[atom] = id
        id_arr_dict[id] = atom

    bonded_lst = [
        (id, [arr_id_dict[a] for a in mg.bonded_to(atom)])
        for atom, id in arr_id_dict.items()
    ]

    bonded_lst.sort(key=lambda x: len(x[1]))

    id_nbrs_tuple_list = []
    for n_nbrs, group in itertools.groupby(
        bonded_lst, key=lambda x: len(x[1])
    ):
        if n_nbrs == 0:
            continue  # Skip aggregation if no neighbors

        group = list(group)
        ids_lst, nbrs_lists = [], []
        for id, nbrs in group:
            ids_lst.append(id)
            nbrs_lists.append(nbrs)

        ids = np.array(ids_lst, dtype=np.int16)
        nbrs = np.array(nbrs_lists, dtype=np.int16)
        id_nbrs_tuple_list.append((ids, nbrs))

    for _ in itertools.repeat(None):
        for ids, nbrs in id_nbrs_tuple_list:
            # Compute the new hash for each atom based on its neighbors
            atom_hash[ids] = numpy_int_multiset_hash(atom_hash[nbrs])
        atom_hash_view = atom_hash.view()
        atom_hash_view.setflags(write=False)
        yield atom_hash_view


def stereo_morgan_generator(
    smg: StereoMolGraph,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> Iterator[np.ndarray[tuple[int], np.dtype[np.int64]]]:
    n_atoms = len(smg.atoms)
    if atom_labels is not None:
        assert len(atom_labels) == n_atoms

    init_atom_hash = (label_hash(smg, ("atom_type",))
                      if atom_labels is None else atom_labels)
    atom_hash = np.append(init_atom_hash, 0) # 0 for "None" in Stereo.atoms

    atom_hash_view = atom_hash.view()
    atom_hash_view.setflags(write=False)
    yield atom_hash_view

    if len(smg.bonds) == 0:
        while True:
            atom_hash_view = atom_hash.view()
            atom_hash_view.setflags(write=False)
            yield atom_hash_view

    prev_atom_hash = np.zeros_like(atom_hash, dtype=np.int64)

    arr_id_dict, id_arr_dict = {}, {}
    stereo_hash_pointer = {}
    for id, atom in enumerate(smg.atoms):
        arr_id_dict[atom] = id
        id_arr_dict[id] = atom
        stereo_hash_pointer[id] = []  # arr_id: list[stereo_pointer]

    arr_id_dict[None] = -1
    id_arr_dict[-1] = None

    grouped_atom_stereo: dict = defaultdict(list)
    atoms_with_atom_stereo: set[int] = set()
    atoms_with_bond_stereo: set[int] = set()

    grouped_bond_stereo: dict = defaultdict(list)
    atoms_with_atom_stereo: set[int] = set()

    as_atoms = []
    as_perm_atoms = []
    as_nbr_atoms = []

    bs_atoms = []
    bs_nbr_atoms = []
    bs_perm_atoms = []

    # i: arrays to store intermediate values.
    # Avoids additional memory allocation.
    i_a_perm_nbrs = []
    i_a_perm = []
    i_a = []

    i_b_perm_nbrs = []
    i_b_perm = []
    i_b = []

    for atom, stereo in smg.atom_stereo.items():
        if stereo.parity is not None:
            nbr_atoms = (
                stereo.atoms
                if stereo.parity != -1
                else stereo._inverted_atoms()  # type: ignore
            )

            grouped_atom_stereo[stereo.PERMUTATION_GROUP].append(
                (atom, nbr_atoms)
            )

            atoms_with_atom_stereo.add(atom)

    for bond, stereo in smg.bond_stereo.items():
        if stereo.parity is not None:
            nbr_atoms = (
                stereo.atoms
                if stereo.parity != -1
                else stereo._inverted_atoms()  # type: ignore
            )

            grouped_bond_stereo[stereo.PERMUTATION_GROUP].append(
                (bond, nbr_atoms)
            )

            for atom in bond:
                atoms_with_bond_stereo.add(atom)

    atoms_without_atom_stereo = set(smg.atoms) - atoms_with_atom_stereo

    for atom in atoms_without_atom_stereo:
        fake_stereo_atoms = (atom, *smg.bonded_to(atom))
        perm_gen = itertools.permutations(range(1, len(fake_stereo_atoms)))
        perm_group = tuple((0, *perm) for perm in perm_gen)
        grouped_atom_stereo[perm_group].append((atom, fake_stereo_atoms))

    # atom_stereo
    for perm_group, atom_nbr_atoms_list_tup in grouped_atom_stereo.items():
        atom_arr_ids = np.array(
            [arr_id_dict[atom] for atom, _ in atom_nbr_atoms_list_tup],
            dtype=np.int16,
        )
        as_atoms.append(atom_arr_ids)

        nbr_atoms = np.array(
            [
                [arr_id_dict[a] for a in nbr_lst]
                for _atom, nbr_lst in atom_nbr_atoms_list_tup
            ],
            dtype=np.int16,
        )

        as_nbr_atoms.append(nbr_atoms)

        perm_group = np.array(perm_group, dtype=np.int8)
        perm_atoms = nbr_atoms[..., perm_group]
        as_perm_atoms.append(perm_atoms)

        # intermediate arrays
        a_perm_nbrs = np.zeros(perm_atoms.shape, dtype=np.int64)
        i_a_perm_nbrs.append(a_perm_nbrs)
        i_a_perm.append(np.zeros(a_perm_nbrs.shape[0:2], dtype=np.int64))
        i_atom_stereo = np.zeros(a_perm_nbrs.shape[0:1], dtype=np.int64)
        i_a.append(i_atom_stereo)

        for stereo_id, atom_arr_id in enumerate(atom_arr_ids):
            stereo_hash_pointer[atom_arr_id].append(
                i_atom_stereo[stereo_id : stereo_id + 1]
            )
            # by reference

    # bond_stereo
    for perm_group, atom_nbr_atoms_list_tup in grouped_bond_stereo.items():
        atom_arr_ids = np.array(
            [
                [arr_id_dict[atom] for atom in bond]
                for bond, _ in atom_nbr_atoms_list_tup
            ],
            dtype=np.int16,
        )
        bs_atoms.append(atom_arr_ids)

        nbr_atoms = np.array(
            [
                [arr_id_dict[a] for a in nbr_lst]
                for _atom, nbr_lst in atom_nbr_atoms_list_tup
            ],
            dtype=np.int16,
        )

        bs_nbr_atoms.append(nbr_atoms)
        arr_perm_group = np.array(perm_group, dtype=np.int8)
        perm_atoms = nbr_atoms[..., arr_perm_group]
        bs_perm_atoms.append(perm_atoms)

        # intermediate arrays
        b_perm_nbrs = np.zeros(perm_atoms.shape, dtype=np.int64)
        i_b_perm_nbrs.append(b_perm_nbrs)
        i_b_perm.append(np.zeros(b_perm_nbrs.shape[0:2], dtype=np.int64))
        i_bond_stereo = np.zeros(b_perm_nbrs.shape[0:1], dtype=np.int64)
        i_bond_stereo.fill(numpy_int_tuple_hash(numpy_int_tuple_hash(arr_perm_group)))
        i_b.append(i_bond_stereo)

        for stereo_id, (atom_arr_id1, atom_arr_id2) in enumerate(atom_arr_ids):
            stereo_hash_pointer[atom_arr_id1].append(
                i_bond_stereo[stereo_id : stereo_id + 1]
            )
            stereo_hash_pointer[atom_arr_id2].append(
                i_bond_stereo[stereo_id : stereo_id + 1]
            )
        # by reference

    i_atoms_with_n_stereo = []  # atoms, i_stereo, group

    pntr = sorted(
        ((id, ptr) for id, ptr in stereo_hash_pointer.items() if ptr),
        key=lambda x: len(x[1]),
    )

    for key, group in itertools.groupby(pntr, key=lambda x: len(x[1])):
        group = list(group)

        ids = []
        pntr_groups = []
        for id, ptrs in group:
            ids.append(id)
            pntr_groups.append(ptrs)

        atoms = np.array(ids, dtype=np.int16)
        i_hash = np.empty((len(pntr_groups), key), dtype=np.int64)
        i_atoms_with_n_stereo.append((atoms, i_hash, pntr_groups))

    for count in itertools.count(1, 1):
        # atom stereo
        for perm_atoms, a_perm_nbrs, a_perm, a in zip(
            as_perm_atoms, i_a_perm_nbrs, i_a_perm, i_a
        ):
            numpy_int_tuple_hash(atom_hash[perm_atoms], out=a_perm)
            numpy_int_multiset_hash(a_perm, out=a)

        # bond stereo
        if count != 0:
            for perm_atoms, b_perm_nbrs, b_perm, b in zip(
                bs_perm_atoms, i_b_perm_nbrs, i_b_perm, i_b
            ):
                numpy_int_tuple_hash(prev_atom_hash[perm_atoms], out=b_perm)
                numpy_int_multiset_hash(b_perm, out=b)

        for (
            atoms,
            i_stereo,
            ptr_lsts,
        ) in i_atoms_with_n_stereo:  # atoms, i_stereo, group
            prev_atom_hash[:] = atom_hash[:]

            i_stereo[:] = np.asarray(
                [[ptr.item() for ptr in ptr_list] for ptr_list in ptr_lsts]
            )
            atom_hash[atoms] = numpy_int_multiset_hash(i_stereo)
        atom_hash_view = atom_hash.view()
        atom_hash_view.setflags(write=False)
        yield atom_hash_view


def _reaction_generator(
    graph: CondensedReactionGraph,
    generator: Callable,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> Iterator[np.ndarray[tuple[int], np.dtype[np.int64]]]:
    color_iters = [
        generator(graph.reactant(), atom_labels=atom_labels),
        generator(graph.product(), atom_labels=atom_labels),
        generator(graph._ts(), atom_labels=atom_labels),
    ]

    stacked: np.ndarray | None = None
    hash_buf: np.ndarray | None = None

    while True:
        for axis, it in enumerate(color_iters):
            color = next(it)
            if stacked is None:
                stacked = np.empty((*color.shape, len(color_iters)),
                                   dtype=np.int64)
                hash_buf = np.empty(color.shape, dtype=np.int64)
            np.copyto(stacked[..., axis], color)

        assert stacked is not None and hash_buf is not None
        hashed = numpy_int_tuple_hash(stacked, out=hash_buf)
        hash_view = hashed.view()
        hash_view.setflags(write=False)
        yield hash_view


def reaction_morgan_generator(
    graph: CondensedReactionGraph,
    max_iter: int | None = None,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> Iterator[np.ndarray[tuple[int], np.dtype[np.int64]]]:
    return _reaction_generator(
        graph=graph,
        generator=morgan_generator,
        atom_labels=atom_labels,
    )


def stereo_reaction_morgan_generator(
    graph: StereoCondensedReactionGraph,
    max_iter: int | None = None,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> Iterator[np.ndarray[tuple[int], np.dtype[np.int64]]]:
    return _reaction_generator(
        graph=graph,
        generator=stereo_morgan_generator,
        atom_labels=atom_labels,
    )


def _color_refine(
    graph: MolGraph,
    generator: Callable,
    max_iter: int | None = None,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> np.ndarray[tuple[int], np.dtype[np.int64]]:
    sm_generator = generator(graph, atom_labels=atom_labels)

    n_atoms = graph.n_atoms

    atom_hash = next(sm_generator)
    n_atom_classes = np.unique(atom_hash).shape[0]

    counter = (
        itertools.count(1, 1) if max_iter is None else range(max_iter + 1)
    )
    for _ in counter:
        atom_hash = next(sm_generator)
        new_n_classes = np.unique(atom_hash).shape[0]
        if new_n_classes == n_atom_classes:
            break
        elif new_n_classes == n_atoms:
            break
        else:
            n_atom_classes = new_n_classes

    return atom_hash


def color_refine_mg(
    graph: MolGraph,
    max_iter: int | None = None,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> np.ndarray[tuple[int], np.dtype[np.int64]]:
    return _color_refine(
        graph=graph,
        generator=morgan_generator,
        max_iter=max_iter,
        atom_labels=atom_labels,
    )


def color_refine_smg(
    graph: StereoMolGraph,
    max_iter: int | None = None,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> np.ndarray[tuple[int], np.dtype[np.int64]]:
    return _color_refine(
        graph=graph,
        generator=stereo_morgan_generator,
        max_iter=max_iter,
        atom_labels=atom_labels,
    )


def color_refine_crg(
    graph: CondensedReactionGraph,
    max_iter: int | None = None,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> np.ndarray[tuple[int], np.dtype[np.int64]]:
    return _color_refine(
        graph=graph,
        generator=reaction_morgan_generator,
        max_iter=max_iter,
        atom_labels=atom_labels,
    )

def color_refine_scrg(
    graph: StereoCondensedReactionGraph,
    max_iter: int | None = None,
    atom_labels: None | np.ndarray[tuple[int], np.dtype[np.int64]] = None,
) -> np.ndarray[tuple[int], np.dtype[np.int64]]:
    return _color_refine(
        graph=graph,
        generator=stereo_reaction_morgan_generator,
        max_iter=max_iter,
        atom_labels=atom_labels,
    )


def color_refine_hash_mg(graph: MolGraph) -> int:
    """Color-refined hash for plain `MolGraph` objects."""
    initial_color_array = np.array(graph.atom_types, dtype=np.int64)
    color_array = color_refine_mg(graph, atom_labels=initial_color_array)
    return int(numpy_int_multiset_hash(color_array))


def color_refine_hash_smg(graph: StereoMolGraph) -> int:
    """Color-refined hash for `StereoMolGraph` objects.

    Drops the extra sentinel slot the stereo generator appends.
    """
    initial_color_array = np.array(graph.atom_types, dtype=np.int64)
    color_array = color_refine_smg(graph, atom_labels=initial_color_array)
    return int(numpy_int_multiset_hash(color_array))


def color_refine_hash_crg(graph: CondensedReactionGraph) -> int:
    """Color-refined hash for `CondensedReactionGraph` objects."""
    color_array = color_refine_crg(graph)
    return int(numpy_int_multiset_hash(color_array))


def color_refine_hash_scrg(graph: StereoCondensedReactionGraph) -> int:
    """Color-refined hash for `StereoCondensedReactionGraph` objects."""
    color_array = color_refine_scrg(graph)
    return int(numpy_int_multiset_hash(color_array))


from __future__ import annotations

import itertools
import json
from collections import deque
from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Any

from stereomolgraph import (
    AtomId,
    Bond,
    CondensedReactionGraph,
    MolGraph,
    StereoCondensedReactionGraph,
    StereoMolGraph,
)
from stereomolgraph.algorithms.color_refine import color_refine_smg
from stereomolgraph.algorithms.isomorphism import vf2pp_all_isomorphisms
from stereomolgraph.graphs.crg import Change
from stereomolgraph.periodic_table import SYMBOLS
from stereomolgraph.stereodescriptors import (
    AtropBond,
    Octahedral,
    PlanarBond,
    SquarePlanar,
    Tetrahedral,
    TrigonalBipyramidal,
)

STEREO_CLASSES: dict[str, type] = {
    "Tetrahedral": Tetrahedral,
    "TrigonalBipyramidal": TrigonalBipyramidal,
    "Octahedral": Octahedral,
    "SquarePlanar": SquarePlanar,
    "PlanarBond": PlanarBond,
    "AtropBond": AtropBond,
}


class JSONHandler:
    """Serialize and deserialize StereoMolGraph-related graphs to JSON."""

    @staticmethod
    def as_dict(graph: MolGraph) -> dict[str, Any]:
        graph_type = type(graph).__name__

        atoms_dict = sorted(
            (
                atom,
                SYMBOLS[a_type],
            )
            for atom, a_type in zip(graph.atoms, graph.atom_types)
        )

        data: dict[str, Any] = {"Atoms": atoms_dict}

        if isinstance(graph, CondensedReactionGraph):
            formed_bonds = graph.get_formed_bonds()
            broken_bonds = graph.get_broken_bonds()

            bonds_dict = sorted(
                sorted(bond)
                for bond in graph.bonds
                if bond not in formed_bonds | broken_bonds
            )
            data["Bonds"] = bonds_dict
            data["Formed Bonds"] = sorted(
                tuple(sorted(bond)) for bond in formed_bonds
            )
            data["Broken Bonds"] = sorted(
                tuple(sorted(bond)) for bond in broken_bonds
            )
        else:
            bonds_dict = sorted(tuple(sorted(bond)) for bond in graph.bonds)
            data["Bonds"] = bonds_dict

        if isinstance(graph, StereoMolGraph):
            atom_stereos: dict[Any, Any] = {}
            bond_stereos: dict[Any, Any] = {}

            for atom, stereo in graph.atom_stereo.items():
                atom_stereos[atom] = {
                    stereo.__class__.__name__: (stereo.atoms, stereo.parity)
                }
            if atom_stereos:
                data["Atom Stereo"] = atom_stereos

            for bond_fset, stereo in graph.bond_stereo.items():
                bond_key = json.dumps(sorted(bond_fset))
                bond_stereos[bond_key] = {
                    stereo.__class__.__name__: (stereo.atoms, stereo.parity)
                }
            if bond_stereos:
                data["Bond Stereo"] = bond_stereos

        if isinstance(graph, StereoCondensedReactionGraph):
            atom_changes: dict[Any, Any] = {}

            for atom, change_dict in graph.atom_stereo_changes.items():
                atom_change_dict: dict[str, Any] = {}
                for change, stereo in change_dict.items():
                    if stereo is not None:
                        atom_change_dict[change.name] = {
                            stereo.__class__.__name__: (
                                stereo.atoms,
                                stereo.parity,
                            )
                        }
                if atom_change_dict:
                    atom_changes[atom] = atom_change_dict
            if atom_changes:
                data["Atom Stereo Changes"] = atom_changes

            bond_changes: dict[Any, Any] = {}
            for bond_fset, change_dict in graph.bond_stereo_changes.items():
                bond_key = json.dumps(sorted(bond_fset))
                bond_change_dict: dict[str, Any] = {}
                for change, stereo in change_dict.items():
                    if stereo is not None:
                        bond_change_dict[change.name] = {
                            stereo.__class__.__name__: (
                                stereo.atoms,
                                stereo.parity,
                            )
                        }
                if bond_change_dict:
                    bond_changes[bond_key] = bond_change_dict
            if bond_changes:
                data["Bond Stereo Changes"] = bond_changes

        return {graph_type: data}

    @classmethod
    def json_serialize(cls, graph: MolGraph) -> str:
        return json.dumps(cls.as_dict(graph))

    @staticmethod
    def _stereo_from_payload(payload: Any):
        if not payload:
            return None
        class_name, (atoms, parity) = next(iter(payload.items()))
        stereo_cls = STEREO_CLASSES[class_name]
        return stereo_cls(tuple(int(a) for a in atoms), parity)

    @classmethod
    def json_deserialize(cls, payload: str) -> MolGraph:
        graph_type, graph_payload = next(iter(json.loads(payload).items()))
        graph_cls = {
            "MolGraph": MolGraph,
            "StereoMolGraph": StereoMolGraph,
            "CondensedReactionGraph": CondensedReactionGraph,
            "StereoCondensedReactionGraph": StereoCondensedReactionGraph,
        }[graph_type]

        graph = graph_cls()

        for atom_id, atom_type in graph_payload.get("Atoms", []):
            graph.add_atom(int(atom_id), atom_type)

        for bond_entry in graph_payload.get("Bonds", []):
            graph.add_bond(*map(int, bond_entry))

        if isinstance(graph, CondensedReactionGraph):
            for bond_entry in graph_payload.get("Formed Bonds", []):
                graph.add_formed_bond(*map(int, bond_entry))
            for bond_entry in graph_payload.get("Broken Bonds", []):
                graph.add_broken_bond(*map(int, bond_entry))

        if isinstance(graph, StereoMolGraph):
            for entry in graph_payload.get("Atom Stereo", {}).values():
                stereo_obj = cls._stereo_from_payload(entry)
                if stereo_obj is not None:
                    graph.set_atom_stereo(stereo_obj)

            for entry in graph_payload.get("Bond Stereo", {}).values():
                stereo_obj = cls._stereo_from_payload(entry)
                if stereo_obj is not None:
                    graph.set_bond_stereo(stereo_obj)

        if isinstance(graph, StereoCondensedReactionGraph):
            for change_dict in graph_payload.get(
                "Atom Stereo Changes", {}
            ).values():
                broken = cls._stereo_from_payload(change_dict.get("BROKEN"))
                formed = cls._stereo_from_payload(change_dict.get("FORMED"))
                fleeting = cls._stereo_from_payload(
                    change_dict.get("FLEETING")
                )
                if any((broken, formed, fleeting)):
                    graph.set_atom_stereo_change(
                        broken=broken,
                        formed=formed,
                        fleeting=fleeting,
                    )

            for change_dict in graph_payload.get(
                "Bond Stereo Changes", {}
            ).values():
                broken = cls._stereo_from_payload(change_dict.get("BROKEN"))
                formed = cls._stereo_from_payload(change_dict.get("FORMED"))
                fleeting = cls._stereo_from_payload(
                    change_dict.get("FLEETING")
                )
                if any((broken, formed, fleeting)):
                    graph.set_bond_stereo_change(
                        broken=broken,
                        formed=formed,
                        fleeting=fleeting,
                    )

        return graph


if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from stereomolgraph.graphs.smg import AtomId, Bond, StereoMolGraph


def unique_generator(input_generator: Iterator) -> Iterator:
    """
    A generator that yields unique objects from another generator.

    Args:
        input_generator: A generator yielding hashable objects

    Yields:
        Only the first occurrence of each unique object from the input generator
    """
    seen_hash = set()
    for item in input_generator:
        item_hash = hash(item)
        if item_hash not in seen_hash:
            seen_hash.add(item_hash)
            yield item


def generate_stereoisomers(
    graph: StereoMolGraph,
    enantiomers: bool = True,
    atoms: None | Iterable[AtomId] = None,
    bonds: None | Iterable[Bond] = None,
) -> Iterator[StereoMolGraph]:
    """Generates all unique stereoisomers of a StereoMolGraph by generation of
    all combinations of parities. Only includes stereocenters which have a
    parity of None. If a parity is set, it is not changed.

    If include_enantiomers is True, both enantiomers of a stereoisomer are
    included, if it is False, only one enantiomer is included.

    Args:
        enantiomers: If True, both enantiomers are included
        atoms: Optional subset of atoms to consider for stereoisomerism
        bonds: Optional subset of bonds to consider for stereoisomerism

    Yields:
        StereoMolGraph: Each unique stereoisomer (and enantiomer if requested)
    """
    if atoms is None:
        atom_stereos = (
            stereo.get_isomers()
            for a in graph.atoms
            if ((stereo := graph.get_atom_stereo(a)) and stereo.parity is None)
        )
    else:
        atom_stereos = (
            stereo.get_isomers()
            for a in atoms
            if (stereo := graph.get_atom_stereo(a)) is not None
        )

    if bonds is None:
        bond_stereos = (
            stereo.get_isomers()
            for b in graph.bonds
            if ((stereo := graph.get_bond_stereo(b)) and stereo.parity is None)
        )
    else:
        bond_stereos = (
            stereo.get_isomers()
            for b in bonds
            if (stereo := graph.get_bond_stereo(b)) is not None
        )

    seen_hash: set[int] = set()
    enantiomers_seen_hash: set[int] = set()

    for a_stereos, b_stereos in itertools.product(
        itertools.product(*atom_stereos), itertools.product(*bond_stereos)
    ):
        stereoisomer = graph.copy()
        for a_stereo in a_stereos:
            stereoisomer.set_atom_stereo(a_stereo)
        for b_stereo in b_stereos:
            stereoisomer.set_bond_stereo(b_stereo)

        stereoisomer_hash = hash(stereoisomer)
        if stereoisomer_hash not in seen_hash:
            seen_hash.add(stereoisomer_hash)
            yield stereoisomer

            if not enantiomers:
                enantiomer = stereoisomer.enantiomer()
                enantiomer_hash = hash(enantiomer)
                if enantiomer_hash not in enantiomers_seen_hash:
                    enantiomers_seen_hash.add(enantiomer_hash)
                    yield enantiomer


def generate_fleeting_stereoisomers(
    graph: StereoCondensedReactionGraph,
    enantiomers: bool = True,
    atoms: None | Iterable[AtomId] = None,
    bonds: None | Iterable[Bond] = None,
) -> Iterator[StereoCondensedReactionGraph]:
    """Generates all unique fleeting stereoisomers of a StereoCondensedReactionGraph.

    Only includes stereocenters which have a parity of None for the fleeting change.
    If a parity is set, it is not changed.

    Args:
        graph: The reaction graph to generate isomers from
        enantiomers: If True, both enantiomers are included (default: True)
        atoms: Optional subset of atoms to consider for stereoisomerism
        bonds: Optional subset of bonds to consider for stereoisomerism

    Yields:
        StereoCondensedReactionGraph: Each unique fleeting stereoisomer
    """
    # Get atom stereoisomers
    if atoms is None:
        atom_stereos = (
            stereo.get_isomers()
            for a in graph.atoms
            if (
                (stereo_change_dict := graph.get_atom_stereo_change(a))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        )
    else:
        atom_stereos = (
            stereo.get_isomers()
            for a in atoms
            if (
                (stereo_change_dict := graph.get_atom_stereo_change(a))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        )

    # Get bond stereoisomers
    if bonds is None:
        bond_stereos = (
            stereo.get_isomers()
            for b in graph.bonds
            if (
                (stereo_change_dict := graph.get_bond_stereo_change(b))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        )
    else:
        bond_stereos = (
            stereo.get_isomers()
            for b in bonds
            if (
                (stereo_change_dict := graph.get_bond_stereo_change(b))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        )

    seen_isomers_hash = set()
    seen_enantiomers_hash = set()

    for a_stereos, b_stereos in itertools.product(
        itertools.product(*atom_stereos), itertools.product(*bond_stereos)
    ):
        stereoisomer = graph.copy()
        for a_stereo in a_stereos:
            stereoisomer.set_atom_stereo_change(fleeting=a_stereo)
        for b_stereo in b_stereos:
            stereoisomer.set_bond_stereo_change(fleeting=b_stereo)

        stereoisomer_hash = hash(stereoisomer)
        if stereoisomer_hash not in seen_isomers_hash:
            seen_isomers_hash.add(stereoisomer_hash)
            yield stereoisomer

            if not enantiomers:
                enantiomer = stereoisomer.enantiomer()
                enantiomer_hash = hash(enantiomer)
                if enantiomer_hash not in seen_enantiomers_hash:
                    seen_enantiomers_hash.add(enantiomer_hash)
                    yield enantiomer


def topological_symmetry_number(graph: StereoMolGraph) -> int:
    """
    Calculated from the number of graph isomorphisms which conserve the
    stereo information.
    symmetry_number = internal_symmetry_number * rotational_symmetry_number
    TODO: add paper reference
    """

    if any(stereo.parity is None for stereo in graph.stereo.values()):
        raise NotImplementedError(
            "all stereocenters have to be defined"
            " to calculate the symmetry number"
        )
    colorings = color_refine_smg(graph)
    mappings = vf2pp_all_isomorphisms(
        graph, graph, atom_labels=(colorings, colorings), stereo=True
    )
    return deque(enumerate(mappings, 1), maxlen=1)[0][0]
